#
# Virtual Traces Experiment Software Distribution.
# This code is licensed under GPL v2, and is copyright (C) Jack Whitham 2008.
# See PATCH_NOTE.
#

import pickle, cycle, math, glob, os

RESULTS = pickle.load(file("results_data"))
PROGRAM_LIST = pickle.load(file("bench_data"))
PROGRAM_LIST.sort()
L_VALUES = cycle.L_VALUES[:]
L_VALUES.sort()
IGNORE_LIST = [ 'somecode' ]
MIN_REF = 1000
OUTDIR = "r1/"


def Escape(name):
    return name.replace('_', r'\_')

def Draw_Table(table_name, key_list, size=r"\small", be_clever=False):
    out_csv = []
    out_tex = [ r'\begin{tabular}{|c|' ]
    max_h = 0
    for (headings, key) in key_list:
        out_tex.append('r|')
        assert type(headings) == list
        max_h = max(len(headings), max_h)
    out_tex.append('}\n')
    out_tex.append(r'\hline')
    out_tex.append('\n')
    out_tex.append('{%s Program}' % size)
    for h in xrange(max_h):
        line = []
        for (headings, key) in key_list:
            if ( h < len(headings) ):
                line.append((headings[ h ], 1))
            else:
                line.append(('', 1))

            if ( be_clever and len(line) > 1 ):
                (heading2, count2) = line[ -2 ]
                (heading1, count1) = line[ -1 ]
                if ( heading1 == heading2 ):
                    count2 += 1
                    line[ -2 ] = (heading2, count2)
                    line.pop()

        for (heading, count) in line:
            out_tex.append(' & ')
            out_tex.append(r'\multicolumn{%u}{c|}{{%s %s}} ' % (
                                count, size, heading))

        out_tex.append(r' \\')
        out_tex.append('\n')

    out_tex.append(r'\hline')
    out_tex.append('\n')

    for record in PROGRAM_LIST:
        base_name = record[ 'base_name' ]
        if ( base_name in IGNORE_LIST ):
            continue

        out_csv.append('%20s' % (base_name))
        out_tex.append('{%s %s}' % (size, Escape(base_name)))
        for (headings, key) in key_list:
            k = tuple([ base_name ] + list(key))
            val = RESULTS.get(k, None)
            if ( type(val) in (int, float, long) ):
                out_csv.append(', %7u' % val)
                out_tex.append(' & {%s %u}' % (size, val))
            else:
                print 'not found', k
                out_csv.append(', ')
                out_tex.append(' & ')
        out_csv.append('\n')
        out_tex.append(r' \\')
        out_tex.append('\n')

    out_tex.append(r'\hline\end{tabular}')
    out_tex.append('\n\n\n')
    
    file(OUTDIR + table_name + '.csv', 'wt').write(''.join(out_csv))
    file(OUTDIR + table_name + '.tex', 'wt').write(''.join(out_tex))

def Exit_Data():
    out = file(OUTDIR + 'exit_data.csv', 'wt')
    L = max(L_VALUES)

    for record in PROGRAM_LIST:
        base_name = record[ 'base_name' ]
        if ( base_name in IGNORE_LIST ):
            continue
        
        o3vtc_time = RESULTS.get((base_name, 'timing', L))
        iio_time = RESULTS.get((base_name, 'iio_cost', 1))
        time_ratio = float(o3vtc_time) / float(iio_time)

        trace_count = RESULTS.get((base_name, 'trace_count', L))
        exit_count = RESULTS.get((base_name, 'exit_count', L))
        exit_ratio = float(exit_count) / float(trace_count)

        out.write('%s,%1.5f,%1.5f\n' % (base_name,
                time_ratio, exit_ratio))

    out.close()


class Investigate_Break_Even:
    def __init__(self):
        pass

    def Run(self, search_dir):
        events = set([ 'JMP', 'BTK', 'BNT', 'END' ])

        for name in glob.glob('%s/*_cmp.txt' % search_dir):
            index = 0

            for line in file(name):
                line = line.rstrip()
                fields = line.split()
                cmd = fields[ 0 ]
                param = int(fields[ -1 ], 16)

                if ( cmd == 'STA' ):
                    self.Start_VT(name, param)
                    index = 0
                elif ( cmd in events ):
                    o3vtc_cost = int(fields[ -2 ], 16)
                    iio_cost = param
                    self.Exit_VT(o3vtc_cost, iio_cost)
                    index += 1

    def Start_VT(self, name, pc):
        pass

    def Exit_VT(self, o3vtc_cost, iio_cost):
        pass

class Get_Specific_Traces(Investigate_Break_Even):
    # Generates somecodeex and contrasting tables.
    def __init__(self):
        self.somecode = []
        self.contrasting = []
        self.collect = None

    def Start_VT(self, name, pc):
        self.collect = None

        for (collect, condition) in [
                (self.somecode, (( name.endswith("somecode_cmp.txt") )
                    and ( pc == 0x120000440 ))),
                (self.contrasting, (( name.endswith("ndes_cmp.txt") )
                    and ( pc == 0x120000d40 ))) ]:

            if ( condition ):
                self.collect = collect

    def Exit_VT(self, o3vtc_cost, iio_cost):
        if ( self.collect != None ):
            self.collect.append((o3vtc_cost, iio_cost))

    def Dump_Tables(self):
        self.Dump("somecodeex", [ '\(i\)', r'\(t_i\)' ], 
                    [ (x,) for (x, y) in self.somecode ], 4)
        self.Dump("contrasting", [ ('', '\(i\)'),
                    (r'\(t_i\)', r'{\hz O3+}', r'{\hz VTC}'), 
                    (r'\(t_i\)', r'{\hz IIO}') ], 
                    self.contrasting, 3)

    def Dump(self, out_name, headings, row_data, horiz, size=r'\small'):
        out_tex = [ r'\begin{tabular}{' ]

        hsize = 1
        for x in xrange(horiz):
            out_tex.append('|')
            for h in headings:
                out_tex.append('c|')
                if ( type(h) in (tuple, list) ):
                    hsize = max(len(h), hsize)

        out_tex.append('}\n')
        out_tex.append(r'\hline')
        out_tex.append('\n')

        for y in xrange(hsize):
            first = True
            for x in xrange(horiz):
                for h in headings:
                    if ( not first ):
                        out_tex.append(' & ')

                    if ( type(h) in (tuple, list) ):
                        if ( y < len(h) ):
                            h = h[ y ]
                        else:
                            h = ''
                    else:
                        if ( y != 0 ):
                            h = ''
            
                    out_tex.append('{{%s %s}}' % (size, h))
                    first = False
            out_tex.append(r'\\' + '\n')

        out_tex.append(r'\hline' + '\n')
        num_rows = ( len(row_data) + horiz - 1 ) / horiz

        for y in xrange(num_rows):
            first = True
            for x in xrange(horiz):
                cell = ( x * num_rows ) + y

                i = 0
                for h in headings:
                    if ( not first ):
                        out_tex.append(' & ')

                    if ( cell < len(row_data) ):
                        if ( i == 0 ):
                            val = '%u' % (( cell + 1 ) % len(row_data))
                        else:
                            val = row_data[ cell ][ i - 1 ]
                    else:
                        val = ''

                    out_tex.append('{%s %s}' % (size, val))
                    first = False
                    i += 1

            out_tex.append(r'\\' + '\n')

        out_tex.append(r'\hline\end{tabular}')
        out_tex.append('\n\n\n')
        
        file(OUTDIR + out_name + '.tex', 'wt').write(''.join(out_tex))

class Get_BE_Proportion(Investigate_Break_Even):
    # Difficult - need (get this...) the number of clock cycles spent in 
    # trace paths that didn't break even. Requires knowledge
    # of IIO values (got that..) but also the path through the code,
    # which is known in capture_metas but is it stored anywhere??
    # _ccm.txt? 
    # For every program.
    def __init__(self):
        self.somecode = []
        self.contrasting = []
        self.collect = None

    def Start_VT(self, name, pc):
        pass

def Make_Graph(out_name, title, key, out_fn):
    local_list = []
    for record in PROGRAM_LIST:
        base_name = record[ 'base_name' ]
        if ( ( not ( base_name in IGNORE_LIST ))
        and ( RESULTS.get((base_name, 'iio_cost', 1), 0) >= MIN_REF )):
            local_list.append(base_name)

    box = file(OUTDIR + out_name + '.dat',  'wt')
    for L in L_VALUES:
        box.write('%-2u' % L)
        for base_name in local_list:
            box.write(' %1.3f' % out_fn(RESULTS.get((base_name, key, L)), base_name))
        box.write('\n')

    box.close()
    box = file(OUTDIR + out_name + '.gnuplot', 'wt')
    box.write(r"""
set terminal postscript landscape monochrome "Times" 12""")
    box.write(r"""
set ylabel "%s" """ % title)
    box.write(r"""
set xlabel "Value of L"
set key box outside right 
set xrange [ 0.5 : 20.5 ]
set size 0.9, 0.54
plot \
""")

    x = 2
    for base_name in local_list:
        if ( x != 2 ):
            box.write(', \\\n')

        box.write('"%s%s.dat" using 1:%u title "%s" with linespoints' % (
                OUTDIR, out_name, x, base_name))
        x += 1
    box.write("\n\n\n")
    box.close()
    rc = os.system('gnuplot %s%s.gnuplot >%s%s.ps' % (OUTDIR, out_name, OUTDIR, out_name))
    assert rc == 0


def Make_Boxgraph():
    Make_Graph('boxgraph', 
        r"Proportion of executed\npaths that break even",
        'be_proportion',
        lambda val, base_name: val)

def Make_Timing():
    Make_Graph('timinggraph', 
        "Normalised execution time",
        'timing',
        lambda val, base_name: (float(val) / 
                float(RESULTS.get((base_name, 'inst_count', 1))))
        )

def Make_Histogram(againstdbp_file, irregular_list):
    divide = 10.0
    irregular_count = []
    for i in xrange(1000):
        irregular_count.append([ 0 for j in irregular_list ])

    irregular_N = [ 0.0 for j in irregular_list ]
    irregular_D = [ 0.0 for j in irregular_list ]

    for line in file(againstdbp_file):
        fields = line.split(',')
        values = [ int(f) for f in fields[ 1: ] ]
        reference = values[ 0 ]
        if ( reference < MIN_REF ):
            print 'reject %s: reference too small' % fields[ 0 ].strip()
            continue

        for (i, j) in enumerate(irregular_list):
            nval = reference / float(values[ i + 1 ])
            inc_box = int(math.floor(0.5 + (nval * divide)))
            inc_box = max(0, min(len(irregular_count) - 1, inc_box))
            irregular_count[ inc_box ][ i ] += 1

            irregular_N[ i ] += nval
            irregular_D[ i ] += 1.0

   
    start = len(irregular_count)
    finish = 0
    for i in xrange(len(irregular_count)):
        for j in irregular_count[ i ]:
            if ( j > 0 ):
                if ( i < start ):
                    start = i
                finish = i

    out = file(OUTDIR + 'histo.csv', 'wt')
    for j in irregular_list:
        out.write(',%s' % j)
    out.write('\n')
    assert start < finish

    for row in xrange(start, finish + 1):
        col = (float(row) / divide)
        out.write('%1.1f' % col)
        for (i, j) in enumerate(irregular_count[ row ]):
            out.write(', %3u' % j)

        out.write('\n')
    out.write('\n')
    out.write('\n')

    for (n, d, j) in zip(irregular_N, irregular_D, irregular_list):
        if ( n != 0.0 ):
            n /= d
        out.write('%s, %1.3f\n' % (j, n))

    out.close()

def Main():
    try:
        os.mkdir(OUTDIR)
    except:
        pass

    o3vtc = r'{\hz O3+VTC}'
    # O3+VTC is better than IIO except for small L.
    Draw_Table("timing", 
        [ ([r'{\hz IIO}'], ('iio_cost', 1)) ] +
            [ ([o3vtc, r'\(L=%u\)' % L], ('timing', L)) for L in [1, 10, 20] ],
            be_clever=True)

    Make_Boxgraph()
    Make_Timing()

    st = Get_Specific_Traces()
    st.Run('L20')
    st.Dump_Tables()

    # Table somecodeex - get it from L20 dir.
    # Table contrasting - same

    # L too short issue. Example: ndes. Graph: X axis L, 
    #   Y1 axis exec time, Y2 axis is trace count.
    Draw_Table("exec_time", 
        [ ([r'{\hz IIO}'], ('iio_cost', 1)) ] +
            [ ([o3vtc, r'\(L=%u\)' % L], ('timing', L)) for L in L_VALUES ],
            be_clever=True)
    Draw_Table("trace_count", 
            [ ([o3vtc + r', \(L=\)', r'\(%u\)' % L], 
            ('trace_count', L)) for L in L_VALUES ],
            be_clever=True)
    Draw_Table("exit_cost", 
            [ ([o3vtc, r'\(L=%u\)' % L], ('exit_count', L)) for L in L_VALUES ],
            be_clever=True)

    # Traces not representative issue
    Exit_Data()

    Draw_Table("againstdbp", 
            [ ([o3vtc, r'\(L=%u\)' % max(L_VALUES)], ('timing', max(L_VALUES))) ] +
            [ ([o3vtc, r'{\hz %s}' % ext_label], ('timing', int_label))
                    for (ext_label, int_label) in [
                            ('-SYNC', 'static1'),
                            ('-SYNC-IOB', 'staticIOB'),
                            ('-SYNC-SD', 'staticSD'),
                            ('-SYNC-NF', 'staticNF') ] ] + [
              ([r'{\hz O3+SBP}'], ('timing', 'static2')),
              ([r'{\hz O3}'], ('timing', 'dbp')) ])

    Make_Histogram(OUTDIR + "againstdbp.csv", [ 'O3+VTC-SYNC', 'O3+VTC-SYNC-IOB',
            'O3+VTC-SYNC-SD', 'O3+VTC-SYNC-NF', 'SBP', 'DBP' ])


if ( __name__ == "__main__" ):
    Main()

