
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include "symboltable.hh"
#include "u64.hh"

Symbol_Table :: Symbol_Table ( FILE * nm )
{
    char        line [ BUFSIZ ] ;

    data_size = 0 ;
    assert ( nm != 0L ) ;
    while ( fgets ( line , BUFSIZ - 1 , nm ) != 0L )
    {
        line [ BUFSIZ - 1 ] = '\0' ;
        int sz = strlen ( line ) ;
        if (( sz > 0x24 )
        && ( ! isspace ( line [ 0 ] )))
        {
            line [ 0x10 ] = '\0' ;
            ulongint address = (ulongint) strtoll ( line , 0L , 16 ) ;
            line [ 0x21 ] = '\0' ;
            ulongint size = (ulongint) strtoll ( & line [ 0x11 ] , 0L , 16 ) ;

            line [ sz - 1 ] = '\0' ;
            const char * name1 = & line [ 0x24 ] ;
            std::string name = name1 ;

            Symbol_Record * r = new Symbol_Record () ;
            r -> address = address ;
            r -> size = size ;
            r -> name = name ;
            table [ address ] = r ;
            reverse_table [ name ] = r ;
            all_records . push_back ( r ) ;

            char sym_type = line [ 0x22 ] ;

            switch ( toupper ( sym_type ))
            {
                case 'B' :
                case 'C' :
                case 'D' :
                case 'G' :
                case 'R' :
                case 'S' :
                        /* data */
                        data_size += size ;
                        break ;
                default :
                        break ;
            }
        }
    }
}

Symbol_Table :: ~Symbol_Table ()
{
    std::vector<Symbol_Record *>::iterator i ;
    
    for ( i = all_records . begin () ; i != all_records . end () ; i ++ )
    {
        Symbol_Record * r = (* i) ;
        delete r ;
    }
}

bool Symbol_Table :: Get_Record ( std::string name ,
            ulongint & address , ulongint & size )
{
    if ( reverse_table . count ( name ) == 0 )
    {
        address = size = 0 ;
        return false ;
    } else {
        Symbol_Record * r = reverse_table [ name ] ;
        address = r -> address ;
        size = r -> size ;
        return true ;
    }
}

void Symbol_Table :: Lookup ( FILE * outfd , ulongint pc )
{
    STMap::iterator i = table . upper_bound ( pc ) ;

    if (( i != table . end () )
    && ( i != table . begin () ))
    {
        i -- ;

        std::string name = (* i) . second -> name ;
        assert ( (* i) . first <= pc ) ;

        fprintf ( outfd , "<%s+%"FU">" , 
                name . c_str() , pc - (* i) . first ) ;
    } else {
        fprintf ( outfd , "<?>" ) ;
    }
}


